/*
 * Copyright (c) 2004-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2004-2006 Stony Brook University
 */
/*
 *  $Id: levels.h,v 1.22 2006/12/13 17:00:21 kolya Exp $
 */

#ifdef __KERNEL__
#ifndef __LEVELS_H_
#define __LEVELS_H_
#include <asm/div64.h>
#include "../raif.h"
#include "../loadbal.h"
#include "../fist.h"
#include "ECC.h"

/* Calculate the number of ECC-dedicated branches */
static inline int ECCWD(int level)
{
	int iECCw = 0;
	if (level == 6)
		iECCw = 2;
	else if (level == 4)
		iECCw = 1;
	return iECCw;
}


/* Calculate the level of redundancy branches */
static inline int ECCW(int level)
{
	int iECCw = 0;
	if (level >= 6)
		iECCw = 2;
	else if (level >= 4)
		iECCw = 1;
	return iECCw;
}

#define CALCULATE_AUTH_RANGE(inode) \
	if (LEVEL(inode) == 1) {\
		imax = bnum;\
		bindex = 0;\
	} else {\
		imax = ECCW(LEVEL(inode)) + 1;\
		bindex = AUTH_BRANCH(inode);\
	}

/* Calculate RAIF file size based on lower files' sizes */
static inline void update_inode_size(inode_t* inode)
{
	int bindex, bnum;
	unsigned i_size = 0;
	unsigned i_size_max = 0;
	unsigned i_blocks_max = 0;
	unsigned i_blocks = 0;
	inode_t* hidden_inode;
	int iECCw;


	if (inode && S_ISREG(inode->i_mode)) {
		bnum = ibnum(inode);
		if (LEVEL(inode) != 1) {

#ifdef RAIF6_ROW_DIAG
			if(LEVEL(inode) == 6)
				bnum--;
#endif
			for (bindex = 0; bindex < bnum; bindex++) {
				hidden_inode = itohi_index(inode, bindex);
#ifdef RAIF_DELAY_CREATE

				if(hidden_inode == NULL)
					continue;

#endif

				i_size += hidden_inode->i_size;
				i_blocks += hidden_inode->i_blocks;
				/* Find max lower file size */
				if (i_size_max < hidden_inode->i_size) {
					i_size_max = hidden_inode->i_size;
					i_blocks_max = hidden_inode->i_blocks;
				}
			}
		} else {
			hidden_inode = itohi_index(inode, 0);
			i_size = hidden_inode->i_size;
			i_blocks = hidden_inode->i_blocks;
		}

		/* Substruct Parity Size */
		iECCw = ECCW(LEVEL(inode));

		fist_dprint(8, "  %s: i_size=%d, i_blocks=%d, i_size_max=%d, i_blocks_max=%d, iECCw=%d\n", __FUNCTION__, i_size, i_blocks, i_size_max, i_blocks_max, iECCw);

#ifdef RAIF6_ROW_DIAG
	if(iECCw == 2)
		iECCw = 1;
#endif

		i_size -= (i_size_max) * iECCw;
		i_blocks -= (i_blocks_max) * iECCw;

		inode->i_size = i_size;
		/* XXX --- calculate i_blocks based on the lower branches */
		if ((i_size & (PAGE_CACHE_SIZE - 1)) == 0)
			i_blocks = i_size >> 9;
		else
			i_blocks = (i_size >> 9) + 8;
		/* XXX --- */
		inode->i_blocks = i_blocks;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)
		inode->i_bytes = i_size - (i_blocks << 9);
#endif

		fist_dprint(8, "  %s: inode: %p, new i_size=%d, i_blocks=%d\n", __FUNCTION__, inode, i_size, i_blocks);
	}
}

/*
 * Calculate the size of the file in the Diagonal Parity Branch
 * in Row-Diagonal Parity Algorithm.
 */

static inline loff_t calculate_rdp_diag_size(struct inode* inode, int bindex, loff_t ia_size)
{
	unsigned int bnum = ibnum(inode);
	unsigned int sush = STRIPING_UNIT_SHIFT(inode);
	unsigned int rdp_array_size = ( bnum - 2 ) * ( bnum - 2 );
	unsigned long strp = ( ia_size >> (PAGE_CACHE_SHIFT + sush) );
	loff_t min_val;
	unsigned int rem;
	
	if(ia_size != 0)
		strp++;

	rem = strp % rdp_array_size;

	min_val = ( (strp - rem) / (bnum - 2) ) << (PAGE_CACHE_SHIFT + sush);

	if(rem == 0)
		return min_val;

	if(rem > bnum - 2)
		return min_val + ( (bnum - 2) << (PAGE_CACHE_SHIFT + sush) );


	return min_val + do_div(ia_size, (rdp_array_size << (PAGE_CACHE_SHIFT + sush) ));
}



/*
 * Calculate lower file size based on RAIF file size 
 * Only setattr uses it so we do not care about performance too much.
 */
static inline loff_t calculate_inode_size(struct inode* inode, int bindex, loff_t ia_size)
{
	int iECCw, iECC, i, imax, bnum, sush, sbr, level;
	unsigned long sunits, top_sunits;
	loff_t low_ia_size = ia_size;

	level = LEVEL(inode);

	/* No striping-related calculations for RAIF1 */
	if (level != 1) {
		bnum = ibnum(inode);
		sush = STRIPING_UNIT_SHIFT(inode);

#ifdef RAIF6_ROW_DIAG
		if(level == 6){
			if(bindex == bnum - 1) {
				return calculate_rdp_diag_size(inode, bindex, ia_size);
			}
			else
				bnum--;
		}
#endif

		sbr = START_BRANCH(inode);
		iECCw = ECCW(level);

#ifdef RAIF6_ROW_DIAG
		if(level == 6)
			iECCw = 1;
#endif

		imax = bnum - iECCw;
		ASSERT(iECCw < bnum);
		/* Number of striping units */
		sunits = ia_size >> (PAGE_SHIFT + sush);
		low_ia_size = ( sunits / imax ) << (PAGE_SHIFT + sush);

		switch (level) {
			case 0:
			case 4:

#ifdef RAIF6_ROW_DIAG
			case 6:
#endif
				iECC = imax;
				i = (bindex + imax - sbr) % imax;
				if (bindex >= iECC)
					i = 0;
				break;
			case 5:

#ifndef RAIF6_ROW_DIAG
			case 6:
#endif
				iECC = (sbr + imax + sunits / imax) % bnum;
				i = (bindex + 2*bnum - (iECC + iECCw)) % bnum;
				if (i >= imax)
					i = 0;
				break;
			default:
				i = iECC = 0;
				BUG();
		} /* switch */

		/* Striping Units In the top stripe */
		top_sunits = (ia_size - low_ia_size * imax) >> (sush + PAGE_SHIFT);

		fist_dprint(8, "%s: i=%d, top_sunits=%d, low_ia_size=%d, ia_size=%d\n", __FUNCTION__, i, top_sunits, low_ia_size, ia_size);
		/* i is in the stripe coordinates (D0, D1, ... Dimax, P1, P2, ... PiECCw) */
		if ( top_sunits > i )
			low_ia_size += PAGE_SIZE << sush;
		else if ( top_sunits == i )
			low_ia_size += ia_size - (sunits << (PAGE_SHIFT + sush));
	}
	return low_ia_size;
}

static inline void COMPUTE_PAGE_LOCATION_STRIPE(inode_t * inode, unsigned page_index, unsigned bnum, unsigned * bindex, unsigned * hidden_page_index)
{
	unsigned sus, sbr, imax, iECCw; 
	sus = STRIPING_UNIT_SIZE(inode);
	sbr = START_BRANCH(inode);
	iECCw = ECCW(LEVEL(inode));
	imax = bnum - iECCw;
	ASSERT(iECCw < bnum);
	switch(LEVEL(inode)) {
		case 0: 
		case 4:
			*bindex = (sbr + page_index / sus) % imax;
			break;
		case 5:
			*bindex = (sbr + (page_index / sus) % imax + iECCw * (page_index / (sus * imax))) % bnum;
			break;
		case 6:
#ifndef RAIF6_ROW_DIAG
			*bindex = (sbr + (page_index / sus) % imax + iECCw * (page_index / (sus * imax))) % bnum;
#else
			*bindex = (sbr + page_index / sus) % imax;
#endif
			break;
		case 1:
			*bindex = sbr;
			break;
		default:
			*bindex = 0;
		ASSERT(0);
	}
	if (LEVEL(inode) != 1)
		*hidden_page_index = (page_index % sus) + sus * ((page_index) / (imax * sus));
	else
		*hidden_page_index = page_index;
}

#define INC_IF_TAG(x) (x)
#define DEC_IF_TAG(x) (x)

static inline void CALCULATE_ECC_BRANCHES(inode_t * inode, unsigned hidden_page_index, unsigned * level, unsigned * iECCw, unsigned * sbr, unsigned * iECC, unsigned * bnum)
{
	unsigned imax, sush;
	*bnum = ibnum(inode);
	*level = LEVEL(inode);
	*iECCw = ECCW(*level);
	sush = STRIPING_UNIT_SHIFT(inode);
	*sbr = START_BRANCH(inode);
	imax = *bnum - *iECCw;
	ASSERT(*iECCw < *bnum);
	if (*level <= 4) {
		*iECC = imax;
	} else {
		hidden_page_index = DEC_IF_TAG(hidden_page_index);
		*iECC = (*sbr + imax + (hidden_page_index >> sush) * (*iECCw)) % (*bnum);
		hidden_page_index = INC_IF_TAG(hidden_page_index);
	}
	fist_dprint(8, "%s: *sbr=%d, *bnum=%d, *level=%d, *iECC=%d, *iECCw=%d\n", __FUNCTION__, *sbr, *bnum, *level, *iECC, *iECCw);
}


static inline void CALCULATE_PAGE_LOCATION(unsigned BL1, page_t *page, inode_t **inode, unsigned *bnum, unsigned *page_index, unsigned * bindex, unsigned * hidden_page_index)
{
	*inode = page->mapping->host;
	*bnum = ibnum(*inode);
	*page_index = page->index;
	if (LEVEL(*inode) == 1) {
		*bindex = (BL1);
		*hidden_page_index = *page_index;
	} else {
		COMPUTE_PAGE_LOCATION_STRIPE(*inode, *page_index, *bnum, bindex, hidden_page_index);
		*bnum = *bindex + 1;
	}
	*hidden_page_index = INC_IF_TAG(*hidden_page_index);
	fist_dprint(7, "Calc-ed page location is [%d:%d]. RAIF%d\n", *bindex, *hidden_page_index, LEVEL(*inode));
}

static inline void CALCULATE_PAGE_LOCATION_READ(page_t *page, inode_t **inode, unsigned *bnum, unsigned *page_index, unsigned * bindex, unsigned * hidden_page_index)
{
#if defined(RAIF_LB) && !defined(NO_LB)
	CALCULATE_PAGE_LOCATION(lb_fastest_branch(page->mapping->host), page, inode, bnum, page_index, bindex, hidden_page_index);
#else
	CALCULATE_PAGE_LOCATION(0, page, inode, bnum, page_index, bindex, hidden_page_index);
#endif
}

static inline void CALCULATE_PAGE_LOCATION_WRITE(page_t *page, inode_t **inode, unsigned *bnum, unsigned *page_index, unsigned * bindex, unsigned * hidden_page_index)
{
	CALCULATE_PAGE_LOCATION(0, page, inode, bnum, page_index, bindex, hidden_page_index);
}

static inline void CALCULATE_DIAG_PAGE_LOCATIONS_RDP(inode_t *inode, unsigned hidden_page_index, unsigned bindex, int *diag1_page_index, int *diag2_page_index)
{
	unsigned imax, sush, level;
	int changed_bindex;

	int diag_no1, diag_no2;
	unsigned sus, strp, strp_inb, bl_num, bnum;

	unsigned iECCw, iECC, sbr;

	sush = STRIPING_UNIT_SHIFT(inode);
	CALCULATE_ECC_BRANCHES (inode, hidden_page_index, &level, &iECCw, &sbr, &iECC, &bnum);  /*sets bnum, level, iECC, iECCw */

	sus = STRIPING_UNIT_SIZE(inode);
	hidden_page_index = DEC_IF_TAG(hidden_page_index);

	strp = hidden_page_index >> sush;

	bl_num = strp / (bnum - 2);
	strp_inb = strp % (bnum - 2);

	imax = bnum - iECCw;
	
	if(bindex >= imax)
		changed_bindex = bindex;
	else
		changed_bindex = (bindex + imax - sbr % imax) % imax;

	diag_no1 = (strp_inb + changed_bindex) % (bnum - 1);
	diag_no2 = (strp_inb + bnum - 2) % (bnum - 1);

	*diag1_page_index = -1;

	if(diag_no1 != (bnum - 2)){
		*diag1_page_index = (bl_num * (bnum - 2) + diag_no1) * sus + hidden_page_index % sus;
		*diag1_page_index = INC_IF_TAG(*diag1_page_index);
	}

	*diag2_page_index = -1;

	if(diag_no2 != (bnum - 2)){
		*diag2_page_index = (bl_num * (bnum - 2) + diag_no2) * sus + hidden_page_index % sus;
		*diag2_page_index = INC_IF_TAG(*diag2_page_index);
	}
}

#endif	/* not __LEVELS_H_ */
#endif /* __KERNEL__ */

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
