/*
 * Copyright (c) 2005-2006 Erez Zadok,
 * Copyright (c) 2005-2006 Nikolai Joukov,
 * Copyright (c) 2005      Arun M. Krishnakumar,
 * Copyright (c) 2006      Chaitanya Patti,
 * Copyright (c) 2005      Sunil Satnur,
 * Copyright (c) 2006      Avishay Traeger,
 * Copyright (c) 2005-2006 Stony Brook University,
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: rules.c,v 1.138 2006/12/13 17:00:21 kolya Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include "fist.h"
#include "raif.h"
#include "ECC/levels.h"

STATIC int raif_rule_load(dentry_t* dparent);
STATIC int raif_rule_store(dentry_t* dparent);

/* A regexp token can be any of the foll */
#define TOKEN_START	0 	/* starts with this */
#define	TOKEN_NORM	1	/* somewhere in the middle */
#define TOKEN_END	2	/* ends with this */

/* returns 0 if string does not match a regexp */
/* returns 1 if string matches a regexp */
/* returns 2 if string == regexp */
int static raif_rule_regexp(char* RegExp, const char* name)
{
	char *regexp = NULL;
	char *tmpregexp;
	char *tmpptr;
	int namelen = strlen(name);
	int reglen;
	int ret = 0;


	if (!RegExp || !(*RegExp)) {
		ret = 0;
		goto out;
	}

	regexp = KMALLOC(strlen(RegExp) + 1, GFP_KERNEL);

	if (!regexp) {
		ret = -ENOMEM;
		goto out;
	}

	tmpregexp = regexp;

	strcpy(tmpregexp, RegExp);
	tmpptr = tmpregexp;
	reglen = strlen(tmpregexp);

	if (*tmpregexp == '*') {
		tmpregexp++;
		reglen --;
	}
	if (!tmpregexp || !(*tmpregexp)) {
		ret = 1;
		goto out_kfree;
	}
	if (!strncmp (tmpregexp, name + namelen - reglen, reglen)) {
		ret = (tmpptr == tmpregexp) ? 2 : 1;
		goto out_kfree;
	}

out_kfree:
	KFREE(regexp);

out:
	return ret;
}

/* this can do some storage and load-balancing */
int static raif_rule_start_branch(dentry_t* dentry)
{
	int i = 0;
	unsigned int sum = 0;
	char* dname = (char*)dentry->d_name.name;
	int dlen = dentry->d_name.len;
	int dlevel = dtopd(dentry)->d.lookup.level;

	for (i = 0, sum = 0; i < dlen ; i++)
		sum += (int)dname[i];
		
	return (unsigned int)sum % (dbnum(dentry) - ECCWD(dlevel));
}


void raif_adjust_for_large_stripe(struct raif_dentry_info* rdi)
{
	unsigned non_par, mask, sbr, nmask;
	int last, index, last_b1;
	switch (rdi->d.lookup.level) {
		case 0:
			rdi->mask = (1 << rdi->d.lookup.start_branch);
			rdi->d.lookup.striping_unit_shift = 0;
			rdi->b_num = 1;
			break;
		case 1:
			rdi->d.lookup.striping_unit_shift = 0;
			break;
		case 4:
			rdi->d.lookup.striping_unit_shift = 0;
			non_par = rdi->b_num - 1;
			mask = rdi->mask;
			rdi->mask = (1 << ( rdi->d.lookup.start_branch % non_par));
			last = -1;
			index = 0;

			while (mask) {
				if (mask & 1)
					last = index;
				mask >>= 1;
				index++;
			}

			if (last == -1 || last <= ( rdi->d.lookup.start_branch % non_par))
				BUG();

			rdi->mask |= (1 << last);
			rdi->b_num = 2;

			break;
		case 5:
			rdi->d.lookup.striping_unit_shift = 0;
			mask = rdi->mask;
			sbr = rdi->d.lookup.start_branch;
			rdi->mask = (1 << sbr);
			last = -1;
			index = 1;

			nmask = mask >> (sbr + 1);

			while (nmask) {
				if (nmask & 1) {
					last = index + sbr;
					break;
				}
				nmask >>= 1;
				index++;
			}

			index = 0;

			if (last == -1 && mask) {
				while (mask && !(mask & 1)) {
					index++;
					mask >>= 1;
				}
				if (index < sbr)
					last = index;
			}
			if (last == -1)
				BUG();

			rdi->mask |= (1 << last);
			rdi->b_num = 2;

			break;

		case 6:
			rdi->d.lookup.striping_unit_shift = 0;
			rdi->d.lookup.level = 1;
			mask = rdi->mask;
			sbr = rdi->d.lookup.start_branch;
			rdi->mask = (1 << sbr);
			last_b1 = -1;
			last = -1;
			index = 1;

			nmask = mask >> (sbr + 1);

			while (nmask) {
				if (nmask & 1) {
					last_b1 = last;
					last = index + sbr;

					if (last_b1 != -1)
						break;
				}
				nmask >>= 1;
				index++;
			}
			
			index = 0;

			if (last_b1 == -1 && mask) {
				while (mask && !(mask & 1)) {
					index++;
					mask >>= 1;
				}
				if (index < sbr)
					last_b1 = index;
			}
			if (last_b1 == -1)
				BUG();

			mask >>= 1;
			index++;

			if (last == -1 && mask) {
				while (mask && !(mask & 1)) {
					index++;
					mask >>= 1;
				}
				if (index < sbr)
					last = index;
			}
			if (last == -1)
				BUG();

			rdi->mask |= (1 << last);
			rdi->mask |= (1 << last_b1);
			rdi->b_num = 3;

			break;

		default:
			rdi->d.lookup.striping_unit_shift = 0;
			break;
	}
}


int raif_rule_remove_special(dentry_t * dentry)
{
	struct raif_dentry_info * rdi_p = dtopd(dentry->d_parent);
	struct rule * rp = NULL, * prev = NULL;
	int match = 0, err = 0;
	write_lock(&(rdi_p->d.rules.spin));
	

	prev = rdi_p->d.rules.local_rule;
	rp = prev;
	
	while (rp) {
		if (rp->special) {
			if (raif_rule_regexp(rp->regexp, dentry->d_name.name) == 2) {
				match = 1; 
				break;
			}
		}
		prev = rp;
		rp = rp->next;
	}

	if (match) {
		if (prev == rp)
			rdi_p->d.rules.local_rule = rp->next;
		else
			prev->next = rp->next; 
		KFREE(rp);
	}
	

	write_unlock(&(rdi_p->d.rules.spin));

	if (match)
		err = raif_rule_store(dentry->d_parent);

	return err;
}

/* This function assumes that the file does not yet exist */
int raif_rule_lookup(dentry_t* dentry)
{
	struct raif_dentry_info* rdi   = dtopd(dentry);
	struct raif_dentry_info* rdi_p = dtopd(dentry->d_parent);
	dentry_t * current_dentry = NULL;
	struct rule* rp = NULL;
	int found = 0;
	int only_local = 0;
	int reg_res = 0, err = 0;

	read_lock(&rdi_p->d.rules.spin);
	rp = rdi_p->d.rules.local_rule;

	/* if no rule matches use replication */
	rdi->mask = rdi->d.lookup.start_branch = rdi->d.lookup.striping_unit_shift = 0x00;
	rdi->d.lookup.level = 1;
	
	if (!strncmp(dentry->d_name.name, RAIF_TMPREFIX, RAIF_TMPREFIX_LEN)) {
		only_local = 1;
		rp = rdi_p->d.rules.rule;
	}

	for (rp = rdi_p->d.rules.local_rule; rp != NULL; rp = rp->next) {
		if ((reg_res = raif_rule_regexp(rp->regexp, dentry->d_name.name)) > 0) {
			rdi->b_num = rp->b_num;
			rdi->d.lookup.level = rp->level;
			rdi->d.lookup.striping_unit_shift = rp->striping_unit_shift;
			if (rp->start_branch == -1)
				rdi->d.lookup.start_branch = raif_rule_start_branch(dentry);
			else
				rdi->d.lookup.start_branch = rp->start_branch;
			rdi->mask = rp->mask;
			found = 1;
			break;
		}
	}

	read_unlock(&rdi_p->d.rules.spin);

	if(found || only_local)
		goto out_default; 

	current_dentry = dentry;

	while (current_dentry->d_parent != current_dentry) {
		rdi_p = dtopd(current_dentry->d_parent);
		read_lock(&rdi_p->d.rules.spin);
		for (rp = rdi_p->d.rules.rule; rp != NULL; rp = rp->next) {
			if ((reg_res = raif_rule_regexp(rp->regexp, dentry->d_name.name)) > 0) {
				rdi->b_num = rp->b_num;
				rdi->d.lookup.level = rp->level;
				rdi->d.lookup.striping_unit_shift = rp->striping_unit_shift;
				if (rp->start_branch == -1)
					rdi->d.lookup.start_branch = raif_rule_start_branch(dentry);
				else
					rdi->d.lookup.start_branch = rp->start_branch;
				rdi->mask = rp->mask;
				found = 1;
				break;
			}
		}
		read_unlock(&rdi_p->d.rules.spin);
		if (found)
			break;
		current_dentry = current_dentry->d_parent;
	}

out_default:
	// Default rule is replication on all branches
	//better to check if the mask is 0x00, as there can
	//be files that do not match any of the rules
	if (!rdi->mask) {
		rdi->b_num =	dbnum(dentry->d_parent);
		rdi->mask =	(1 << rdi->b_num) - 1;
	}

	if(rdi->d.lookup.striping_unit_shift > RAIF_MAX_SUSH)
		raif_adjust_for_large_stripe(rdi);

	return err;
}

/* Here we already know if it is a file or directory */
int raif_rule_interpose(dentry_t* dentry)
{
	int err = 0;
	inode_t* inode = dentry->d_inode; // inode is already checked by the caller
	struct raif_dentry_info* rdi = dtopd(dentry);

	if (S_ISREG(inode->i_mode)) {
		// Copy dentry info to the inode
		// bnum is copied in interpose because it should be done even without rules
		LEVEL(inode) = 			rdi->d.lookup.level;
		STRIPING_UNIT_SHIFT(inode) =	rdi->d.lookup.striping_unit_shift;
		START_BRANCH(inode) =		rdi->d.lookup.start_branch;
	} else {
		rdi->b_num = dbnum(dentry->d_parent);
		rdi->mask = (1 << rdi->b_num) - 1;
		if (S_ISDIR(inode->i_mode)) {
			rwlock_init(&rdi->d.rules.spin);
			// parse rules for this directory and link with parent rules
			err = raif_rule_load(dentry);
		}
	}
	return err;
}

/* Create a new rule for the renamed file / dir */
/* dir not tested yet */
int raif_rule_rename(dentry_t* old_dentry, dentry_t* new_dentry)
{
	int err = 0;

	struct raif_dentry_info *old_rdi = NULL;
	struct raif_dentry_info *new_rdi = NULL;
	struct rule *globalrule = NULL;
	struct rule *localrule = NULL;
	struct rule *localruleprev = NULL;
	int localmatch = 0;
	struct rule *addrule = NULL;

	if (S_ISDIR(new_dentry->d_inode->i_mode)) {

		err = -EPERM;
		/*have to find a way to do this for directories*/
		goto out;

	} else if (S_ISREG(new_dentry->d_inode->i_mode)) {

		addrule = (struct rule *)KMALLOC(sizeof(struct rule) + new_dentry->d_name.len + 1, GFP_KERNEL);
		if (!addrule) {
			err = -ENOMEM;
			goto out;
		}

		old_rdi = dtopd(old_dentry->d_parent);
		new_rdi = dtopd(new_dentry->d_parent);

		write_lock(&old_rdi->d.rules.spin);
		localruleprev = localrule = old_rdi->d.rules.local_rule;
		globalrule = old_rdi->d.rules.rule;
	
		addrule->mask = dtopd(old_dentry)->mask;
		addrule->striping_unit_shift = dtopd(old_dentry)->d.lookup.striping_unit_shift;
		addrule->start_branch = dtopd(new_dentry)->d.lookup.start_branch;
		addrule->level = dtopd(old_dentry)->d.lookup.level;
		strncpy(addrule->regexp, new_dentry->d_name.name, new_dentry->d_name.len);
		addrule->regexp[new_dentry->d_name.len] = 0;
		addrule->b_num = dbnum(old_dentry);
		addrule->special = 1;


		localmatch = 0;
		while (localrule) {
			localmatch = raif_rule_regexp(localrule->regexp, old_dentry->d_name.name);

			if (localmatch < 0) {
				err = localmatch;
				write_unlock(&old_rdi->d.rules.spin);
				goto out;
			}
			if (localmatch == 2 && localrule->special) {
				localmatch = 3;
				break;
			}
			localruleprev = localrule;
			localrule = localrule->next;
		}

		if (localmatch == 3) {
			//delete the old rule here
			if (localruleprev == localrule) {
				old_rdi->d.rules.local_rule = localrule->next;
			} else {
				localruleprev->next = localrule->next;
			}
			KFREE(localrule);
		}

		write_unlock(&old_rdi->d.rules.spin);

		//look for the new name in the global and local rules
		write_lock(&new_rdi->d.rules.spin);

		addrule->next = new_rdi->d.rules.local_rule;
		new_rdi->d.rules.local_rule = addrule;

		write_unlock(&new_rdi->d.rules.spin);

		err = raif_rule_store(new_dentry->d_parent);
	}

out:
	return err;
}

STATIC inline int countbits (int n)
{
	int count;

	for (count = 0; n; count++)
		n = n & (n - 1);

	return count;
}

/* This works for a raif mount with at most 31 branches as of now,
   so bnum must be <= 31 (the MSB is left for the sign to still 
   make sense, so that errors can be distinguished) */
STATIC inline int get_optimal_set(int *sets, int NumOfSets, int bnum)
{
	unsigned int i, bestset = (1 << bnum) - 1;
	int j, count, bestcount = bnum;

	for (i = 1; i < (1 << bnum); i++) {
		for (j = 0; j < NumOfSets; j++) {
			if ((sets[j] & i) == 0)
				break;
		}
		if (j == NumOfSets) {
			count = countbits(i);
			if (count < bestcount) {
				bestset = i;
				bestcount = count;
			}
		}
	}

	return bestset;
}

/* Calculate a minimal subset of branches for readdir */
int raif_rule_readdir(dentry_t* dentry)
{
	// XXX ARUN -- coded. Tested.
	int err = 0;
	unsigned int i = 0;
	struct raif_dentry_info* rdi   = dtopd(dentry);
	struct rule* rp = NULL;
	struct rule *tmptail = NULL;
	struct rule *tmphead = NULL;
	int bnum;
	int *sets = NULL;
	int NumOfSets = 0;
	int delayedlookup = 0;

	print_entry_location();

	PASSERT(dentry);
	ASSERT(S_ISDIR(dentry->d_inode->i_mode));

	if ((err = raif_rule_interpose(dentry))) {
		goto out_unlocked;
	}
	
	bnum = dbnum(dentry);

	rdi = dtopd(dentry);
	PASSERT(rdi);

	read_lock(&rdi->d.rules.spin);

	//make the local continue into the global, and then restore the 
	// local and global lists later.
	tmptail = rdi->d.rules.local_rule;
	if (tmptail) {
		while (tmptail->next != NULL) {
			tmptail = tmptail->next;
		}
		tmptail->next = rdi->d.rules.rule;
		tmphead = rdi->d.rules.local_rule;
	}
	else {
		tmphead = rdi->d.rules.rule;
	}

	rp = tmphead;
	
	if (rp == NULL) {
		// Default is replication
		err = 1;
		goto out;
	}

	delayedlookup = 0;
	NumOfSets = 0;
	for (i = ~0; rp; rp = rp->next) {
		if (rp->striping_unit_shift > RAIF_MAX_SUSH) {
			delayedlookup = 1;
			goto out;
		}
		NumOfSets++;
		i &= rp->mask;
	}

	// readdir one branch
	if (i) {
		for (err = 1; err != (1 << 31); err <<= 1)
			if (err & i)
				break;
		goto out;
	}

	sets = (int *)KMALLOC(NumOfSets * sizeof(int), GFP_KERNEL);
	if (!sets) {
		err = -ENOMEM;
		goto out;
	}

	rp = tmphead;
	for (i = 0; rp && ( i < NumOfSets); i++) {
		sets[i] = rp->mask;
		rp = rp->next;
	}

	NumOfSets = i;

out:
	//now seperate the local and global queues
	if (tmptail) {
		tmptail->next = NULL;
	}

	read_unlock(&rdi->d.rules.spin);

	if (delayedlookup) {
		//lookup all branches
		err = ~0;
		goto out_unlocked;
	} else if (!err)
		err = get_optimal_set(sets, NumOfSets, bnum);
	
	if (sets)
		KFREE((void *)sets);

out_unlocked:
	return err;
}

/* Convert file branch number to global branch number */
int raif_rule_global_bindex(dentry_t* dentry, int bindex)
{
	unsigned int i, mask = dtopd(dentry)->mask;

	// XXX 32 -> sizeof...
	for (i = 0; i < 32; i++) {
		if (mask & 1) {
			if (bindex == 0)
				goto out;
			bindex--;
		}
		mask >>= 1;
	}
	// Illegal bindex for this rule
	BUG();
out:
	return i;
}

/* Convert binary rule to a string GL:POS:REGEXP:SUS:LEVEL:0,1,2 */
char* raif_rule_tostr(struct rule* r, int pos, int global)
{
	char* str = NULL;
	unsigned int i, mask;
	// SUS, LEVEL, and BRANCH are each <= 2 bytes
	int str_len = 5 + 10 + 2 + 2 + 1 ;// branches, commas and \0 here and below

	if (!r)
		goto out;
	
	str_len += strlen(r->regexp) + 3 * r->b_num;

	if (r->special) {
		str_len += 2;
	}

	str = KMALLOC(str_len, GFP_KERNEL);
	if (!str)
		goto out;

	sprintf(str, "%c:%d:%s:%d:%d:", global ? 'G' : 'L', pos, r->regexp, r->striping_unit_shift, r->level);
	mask = r->mask;
	str_len = strlen(str);
	for (i = 0; i < sizeof(int) * 8; i++) {
		if (mask & 1) {
			if (i > 9) {
				str[str_len] = '0' + i/10;
				str_len++;
			}
			str[str_len] = '0' + i%10;
			str_len++;
			str[str_len] = ',';
			str_len++;
		}
		mask >>= 1;
	}
	if (!r->special)
		str[str_len - 1] = 0;
	else if (!global) {
		str[str_len - 1] = ':';
		str[str_len] = '1';
		str_len++;
		str[str_len] = 0;
	}
out:
	return str;
}

int raif_check_branches_rdp(int numbranch)
{
	int branches_allowed[] = {4, 6, 8, 12, 14, 18, 20, 24, 30, 32};
	int no = 10, i;
	for(i = 0; i < no; i++) {
		if(branches_allowed[i] == numbranch)
			return 0;
	}
	printk(KERN_ERR "RAIF_RULES: Invalid number of branches for RDP.\n No. of branches for RDP should be p+1 where p is a prime number greater than 2\n");
	return -EINVAL;
}

/*
 * XXX: this is buggy!
 * Need to check the string length for every token.
 */
STATIC struct rule *raif_rule_parse(char *regexp, int *global, int *position)
{
	char *tmp;
	char *rej_exp = ":"; /* separator in rule string */
	char *rej_br = ","; /* separator in branch list */
	int len, brlen, numbranch = 0, mask = 0, err=0, regexpLen, special = 0, explen;
	unsigned long val;
#define STR_R_GLOBAL	0
#define STR_R_POS	1
#define STR_R_REGEXP	2
#define	STR_R_SHIFT	3
#define STR_R_LEVEL	4
#define	STR_R_BRANCHES	5
#define	STR_R_THIS_DIR	6
	int state = STR_R_GLOBAL;
	struct rule *rule = NULL;
	char RegExp[50] = "";
	char *tmpregexp = RegExp;

	strcpy(tmpregexp, regexp);

	regexpLen = strlen(tmpregexp);
	while (*tmpregexp) {
		len = strcspn(tmpregexp, rej_exp);

		if (!len && state == STR_R_THIS_DIR)
			goto done;

		if (!len) {
			printk(KERN_ERR "Invalid rule string, found a zero length option\n");
			goto out;
		}

		switch (state) {
			case STR_R_GLOBAL:
				if (len != 1) {
					err = -EINVAL;
					tmpregexp[len] = '\0';
					printk(KERN_ERR "Invalid rule option %s\n", tmpregexp);
					goto out;
				}
				switch (*tmpregexp) {
					case 'G':
					case 'g':
						*global = 1;
						break;
					case 'L':
					case 'l':
						*global = 0;
						break;
					default:
						err = -EINVAL;
						printk(KERN_ERR "Invalid global/local option %c\n", *tmpregexp);
						goto out;
				}
				state = STR_R_POS;
				break;

			case STR_R_POS:
				val = simple_strtoul(tmpregexp, &tmp, 10);
				*position = (int) val;
				state = STR_R_REGEXP;
				break;

			case STR_R_REGEXP:
				rule = KMALLOC(sizeof(struct rule) + len + 1, GFP_KERNEL);
				if (!rule) {
					err = -ENOMEM;
					goto out;
				}
				rule->next = NULL;
				strncpy(rule->regexp, tmpregexp, len);
				rule->regexp[len] = '\0';
				state = STR_R_SHIFT;
				break;

			case STR_R_SHIFT:
				val = simple_strtoul(tmpregexp, &tmp, 10);
				/*if (val > 51) {
					printk(KERN_ERR "RAIF_RULES: Striping shift invalid or too big\n");
					goto out;
					}*/
				rule->striping_unit_shift = (unsigned int) val;
				state = STR_R_LEVEL;
				break;
			case STR_R_LEVEL:
                                val = simple_strtoul(tmpregexp, &tmp, 10);
                                if (val != 0 && val != 1 && val != 4 && val != 5) {
#ifdef RAIF6_ROW_DIAG
					if(val == 6){
						rule->level = (unsigned int) val;
						state = STR_R_BRANCHES;
						break;
					}
#endif
                                        printk(KERN_ERR "RAIF_RULES: Invalid RAIF level\n");
					err = -EINVAL;
                                        goto out;
                                }
                                rule->level = (unsigned int) val;
                                state = STR_R_BRANCHES;
                                break;

			case STR_R_BRANCHES:
				while(*tmpregexp && *tmpregexp != rej_exp[0]) {
					brlen = strcspn(tmpregexp, rej_br);
					explen = strcspn(tmpregexp, rej_exp);

					if(explen < brlen) {
						brlen = explen;
					}

					val = simple_strtoul(tmpregexp, &tmp, 10);
					if (val > 31) {
						printk(KERN_ERR "RAIF_RULES: branch number should be less than 32\n");
						goto out;
					}
					if (!(mask & (1 << val))) {
						mask |= ((int)1) << val;
						numbranch++;
					}
					if (*(tmpregexp + brlen) != '\0' && *(tmpregexp + brlen) != rej_exp[0])
						tmpregexp += 1;
					tmpregexp += brlen;
				}
#ifdef RAIF6_ROW_DIAG
				if(rule->level == 6)
					if((err = raif_check_branches_rdp(numbranch)))
						goto out;
#endif
				rule->mask = mask;
				rule->b_num = numbranch;
				state = STR_R_THIS_DIR;
				len = 0;
				break;
			case STR_R_THIS_DIR:
				val = simple_strtoul(tmpregexp, &tmp, 10);
				if (val)
					special = 1;
				goto done;
		}
		tmpregexp += len + 1;
	}
done:
	rule->special = special;
	rule->start_branch = -1;


	return rule;

out:
	if (rule)
		KFREE(rule);
	return ERR_PTR(err);
}

int raif_rule_add(dentry_t *dentry, char *crule, int save_flag)
{
	struct raif_dentry_info *rdi;
	struct raif_dentry_info *rdi_p;
	struct rule *rule, *tmp_rule, *tmp_rule_prev;
	struct rule tmp_rule_head; /* Dummy variable used only for optimization */
	int global, position, tmp_pos;
	int err = 0;

	print_entry_location();

	rdi = dtopd(dentry);
	rdi_p = dtopd(dentry->d_parent);

	rule = raif_rule_parse(crule, &global, &position);
	if (IS_ERR(rule)) {
		err = PTR_ERR(rule);
		goto out;
	}

	write_lock(&rdi->d.rules.spin);

	if (global)
		tmp_rule = rdi->d.rules.rule;
	else 
		tmp_rule = rdi->d.rules.local_rule;

	tmp_rule_prev = &tmp_rule_head;
	tmp_rule_prev->next = tmp_rule;
	tmp_pos = position;
	if (!tmp_rule && position) {
		printk(KERN_ERR "Rule list is empty, try position zero\n");
		goto out_unlock;
	}
	while (tmp_pos && tmp_rule) {
		tmp_rule = tmp_rule->next;
		tmp_rule_prev = tmp_rule_prev->next;
		tmp_pos--;
	}
	if (tmp_pos) {
		printk(KERN_ERR "Cannot insert rule at position %d\n", position);
		printk(KERN_ERR "Position is located outside current directory's rule list\n");
		goto out_unlock;
	}
		
	rule->next = tmp_rule_prev->next;
	tmp_rule_prev->next = rule;

	if (global) {
		rdi->d.rules.rule = tmp_rule_head.next;
	} else {
		rdi->d.rules.local_rule = tmp_rule_head.next;
	}

out_unlock:
	write_unlock(&rdi->d.rules.spin);
	if (err)
		KFREE(rule);
	if (!err && save_flag)
		err = raif_rule_store(dentry);
out:
	print_exit_status(err);
	if (!err) {
		printk("<1>Successfully added rule [%s]\n", crule);
	}
	return err;
}

int raif_rule_del(void)
{
	// XXX SUNIL
	return 0;
}

int raif_rule_list(dentry_t *dentry, struct raif_rulebuf *rulebuf)
{
        int err = 0;
        char *crule, *tmprulebuf;
        struct raif_dentry_info* rdi, *curr_rdi = NULL;
        struct rule* rule;
        int i, len;
	dentry_t * curr_dentry = NULL,  * prev_dentry = NULL;

	print_entry_location();
        tmprulebuf = rulebuf->buf;
        rdi = dtopd(dentry);

        if (rulebuf->ruletype == 1)
                goto list_global;

        if (rulebuf->ruletype != 0) {
                printk(KERN_ERR "Invalid value for ruletype field\n");
                err = -EINVAL;
                goto out;
        }

        if (rulebuf->rulenum == 0) {
                len = strlen("LOCAL RULES:\n") + 1;
                if (tmprulebuf + len > rulebuf->buf + RAIF_RULEBUF_SIZE) {
                        rulebuf->done = 0;
                        goto out;
                }
                snprintf(tmprulebuf, len, "LOCAL RULES:\n");
                tmprulebuf += len - 1;
        }

        read_lock(&rdi->d.rules.spin);
        rule = rdi->d.rules.local_rule;
        for (i = 0; rule && (i < rulebuf->rulenum); rule = rule->next, i++);
        if (!rule)
                goto list_global;
        for (; rule; rule = rule->next, i++) {
                crule = raif_rule_tostr(rule, i, 0);
                rulebuf->rulenum = i;
                len = strlen(crule) + 2;
                if (tmprulebuf + len > rulebuf->buf + RAIF_RULEBUF_SIZE) {
                        KFREE(crule);
                        rulebuf->done = 0;
			read_unlock(&rdi->d.rules.spin);
                        goto out;
                }
                snprintf(tmprulebuf, len, "%s\n", crule);
                tmprulebuf += len - 1;
                KFREE(crule);
        }
	
        rulebuf->ruletype = 1;
        rulebuf->rulenum = 0;
list_global:
	read_unlock(&rdi->d.rules.spin);
        /*rule = rdi->d.rules.rule;*/
        if (rulebuf->rulenum == 0) {
                len = strlen("GLOBAL RULES:\n") + 1;
                if (tmprulebuf + len > rulebuf->buf + RAIF_RULEBUF_SIZE) {
                        rulebuf->done = 0;
                        goto out;
                }
                snprintf(tmprulebuf, len, "GLOBAL RULES:\n");
                tmprulebuf += len - 1;
        }


	i = 0;
	curr_dentry = dentry;
	do {
		curr_rdi = dtopd(curr_dentry);
		read_lock(&curr_rdi->d.rules.spin);

		for (rule = dtopd(curr_dentry)->d.rules.rule; rule && (i < rulebuf->rulenum); rule = rule->next, i++);

		if (rule == NULL) {
			read_unlock(&curr_rdi->d.rules.spin);
			prev_dentry = curr_dentry;
			curr_dentry = curr_dentry->d_parent;
		}
	
	} while ((prev_dentry != curr_dentry) && (i < rulebuf->rulenum));

        if (prev_dentry == curr_dentry) {
                rulebuf->done = 1;
                goto out;
        }

	while ((prev_dentry != curr_dentry)) {
		curr_rdi = dtopd(curr_dentry);
		if (rule == NULL) {
			read_lock(&curr_rdi->d.rules.spin);
			rule = curr_rdi->d.rules.rule;
		}
        	for (; rule; rule = rule->next, i++) {
                	crule = raif_rule_tostr(rule, i, 1);
                	rulebuf->rulenum = i;
               	 	len = strlen(crule) + 2;
                	if (tmprulebuf + len > rulebuf->buf + RAIF_RULEBUF_SIZE) {
                        	KFREE(crule);
                        	rulebuf->done = 0;
                        	goto out;
                	}
                	snprintf(tmprulebuf, len, "%s\n", crule);
                	tmprulebuf += len - 1;
                	KFREE(crule);
        	}

		read_unlock(&curr_rdi->d.rules.spin);
		prev_dentry = curr_dentry;
		curr_dentry = curr_dentry->d_parent;
	}
        rulebuf->done = 1;

out:

        *tmprulebuf = '\0';
        print_exit_status(err);
        return err;

}

/* hidden_dentry is unlocked here */
int raif_rule_unlink_file(dentry_t *hidden_dentry)
{
	int err = 0;
	dentry_t *drule = NULL;
	inode_t * inode = NULL;


	// XXX ARUN check that there are any rules associated with this dentry
	// if not - no lookup is necessary because there should be no rules file.
	drule = lookup_one_len(RAIF_RULES_FILE_NAME, hidden_dentry, RAIF_RULES_FILE_NAME_LEN);
	if (IS_ERR(drule)) {
		err = PTR_ERR(drule);
		goto out;
	}

	if (!drule->d_inode) {
		dput(drule);
		goto out;
	}

	inode = drule->d_inode;

	down(&hidden_dentry->d_inode->i_sem);
	atomic_inc(&inode->i_count);
	err = vfs_unlink(hidden_dentry->d_inode, drule);
	up(&hidden_dentry->d_inode->i_sem);

	dput(drule);

	iput(inode);

	if (err) {
		goto out;
	}

out:
	if (err)
		printk("<0>unlink of [%s] in [%s] returned [%d]\n", hidden_dentry->d_name.name, hidden_dentry->d_parent->d_name.name, err);
	return err;
}

STATIC int raif_rule_read_rule(file_t * file, char *rul_buf)
{
	int err = 0;
	int count = 0;
	off_t retval;
	loff_t res;
	ssize_t readv = 0;
	mm_segment_t oldfs;

	oldfs = set_fs(KERNEL_DS);
	readv = vfs_read(file, rul_buf, RAIF_MAX_RULE_SIZE, &file->f_pos);
	set_fs(oldfs);

	if (readv < 0) {
		err = (int)readv;
		goto out;
	}

	if (readv > 0) {
		while ((count < readv) && (rul_buf[count] != '\n'))
			count++;
		if (count < readv) {
			rul_buf[count] = '\0';
			res = vfs_llseek(file, (count + 1 - readv) , SEEK_CUR);
			retval = res;
			if (res != (loff_t)retval) {
				err = -EOVERFLOW;
				goto out;
			}
		} else {
			err = -EINVAL;
			goto out;
		}
	}
	else if (readv == 0)
		err = 1;
out:
	return err;
}

STATIC int raif_rule_load(dentry_t* dentry)
{
	int		err = 0, err1 = 0;
	dentry_t*	drule;
	file_t          *frule;
	char *rul_buf = NULL;
	struct raif_dentry_info* rdi;
	struct vfsmount * mnt;

	print_entry_location();

	rdi = dtopd(dentry);

	//the rules of the root dentry are to be loaded from a file just
	//during the mount, when the interpose happens; never later!
	if (dentry->d_parent == dentry) {
		if (rdi->d.rules.rule || rdi->d.rules.local_rule)
			goto out;
	} else {
		rdi->d.rules.local_rule = NULL;
		rdi->d.rules.rule = NULL;
	}
	
	drule = lookup_one_len(RAIF_RULES_FILE_NAME, dtohd_index(dentry, 0), RAIF_RULES_FILE_NAME_LEN);
	if (IS_ERR(drule)) {
		err = PTR_ERR(drule);
		goto out;
	}

	if(!drule->d_inode) {
		dput(drule);
		goto out;
	}

	mnt = mntget(stohm_index (dentry->d_sb, 0));
	frule = dentry_open(drule, mnt, O_RDONLY);

	if (IS_ERR(frule)) {
		err = PTR_ERR(frule); 
		goto out;
	}

	rul_buf = (char *) KMALLOC(RAIF_MAX_RULE_SIZE, GFP_KERNEL);

	if (!rul_buf) {
		err = -ENOMEM;
		goto out_fput;
	}

	while ((err1 = raif_rule_read_rule(frule, rul_buf)) == 0) {

		if ((err = raif_rule_add(dentry, rul_buf, 0))) 
			goto out_kfree;
	}

	if (err1 == 1)
		err = 0;
	else
		err = err1;

out_kfree:
	KFREE((void *)rul_buf);

out_fput:
	fput(frule);

out:
	print_exit_status(err);
	return err;
}

STATIC int raif_rule_truncate(dentry_t * dentry)
{
	 int err;
         struct iattr newattrs;
 
         newattrs.ia_size = 0;
         newattrs.ia_valid = ATTR_SIZE | ATTR_CTIME;
 
         down(&dentry->d_inode->i_sem);
         err = notify_change(dentry, &newattrs);
         up(&dentry->d_inode->i_sem);
         return err;
}

STATIC int raif_rule_store(dentry_t* dparent)
{
	int err = 0;
	dentry_t *drule;
	dentry_t *ddir;
	file_t *frule;
	char* crules;
	char* crule;
	struct raif_dentry_info* rdi;
	struct rule* rule;
	int bindex, bnum, buf_len = PAGE_SIZE, rul_len, rul_pos = 0;
	mm_segment_t oldfs;
	struct vfsmount * mnt;
	ssize_t writev;
	struct nameidata nd;

	print_entry_location();

	rdi = dtopd(dparent);
	crules = KMALLOC(buf_len, GFP_KERNEL);
	if (!crules)
		goto out;

	read_lock(&rdi->d.rules.spin);
	/*storing local rules*/
	for (rule = rdi->d.rules.local_rule, bindex = 0; rule; rule = rule->next, bindex++) {
		crule = raif_rule_tostr(rule, bindex, 0);
		if (!crule)
			continue;
		rul_len = strlen(crule);
		if (buf_len < rul_pos + rul_len + 1) {
			// XXX KOLYA realloc the buffer
			BUG();
		}
		strcpy(&crules[rul_pos], crule);
		rul_pos += rul_len + 1;
		crules[rul_pos - 1] = '\n';
		KFREE(crule);
	}


	/*storing global rules*/
	for (rule = rdi->d.rules.rule, bindex = 0; rule; rule = rule->next, bindex++) {

		crule = raif_rule_tostr(rule, bindex, 1);
		if (!crule)
			continue;
		rul_len = strlen(crule);
		if (buf_len < rul_pos + rul_len + 1) {
			// XXX KOLYA realloc the buffer
			BUG();
		}
		strcpy(&crules[rul_pos], crule);
		rul_pos += rul_len + 1;
		crules[rul_pos - 1] = '\n';
		KFREE(crule);
	}
	read_unlock(&rdi->d.rules.spin);


	bnum = dbnum(dparent);
	for (bindex = 0; bindex < bnum; bindex++) {
		ddir = dtohd_index(dparent, bindex);
		drule = lookup_one_len(RAIF_RULES_FILE_NAME, ddir, RAIF_RULES_FILE_NAME_LEN);
		if (IS_ERR(drule)) {
			err = PTR_ERR(drule);
			goto out;
		}
		if (!drule->d_inode) {

			down(&ddir->d_inode->i_sem);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
			err = vfs_create(ddir->d_inode, drule, 0777);
#else
	/*XXX have to check if this is right ... more things to put in nd?*/
			nd.dentry = dget(ddir);
			nd.mnt = mntget(stohm_index(dparent->d_sb, bindex));
			nd.last = ddir->d_name;
			nd.last_type = LAST_NORM;

			err = vfs_create(ddir->d_inode, drule, 0777, &nd);
			
			dput(ddir);
			mntput(nd.mnt);
#endif
			up(&ddir->d_inode->i_sem);
			
			if (err) {
				dput(drule);
				goto out;
			}
		}

		if ((err = raif_rule_truncate(drule))) {
			dput(drule);
			goto out;
		}

		mnt = mntget(stohm_index(dparent->d_sb, bindex));
		frule = dentry_open(drule, mnt, O_WRONLY);
		if (IS_ERR(frule)) {
			err = PTR_ERR(frule);
			goto out;
		}
		if (rul_pos > 0) {
			oldfs = set_fs(KERNEL_DS);
			writev = vfs_write(frule, crules, rul_pos, &frule->f_pos);
			set_fs(oldfs);
			if (writev < 0) {
				err = (int)writev;
				fput(frule);
				goto out;
			}
		}
		fput(frule);
	}
	err = 0;
out:
	if (crules)
		KFREE(crules);
	print_exit_status(err);
	return err;
}

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
